/*
 * Copyright 2019, Cypress Semiconductor Corporation or a subsidiary of
 * Cypress Semiconductor Corporation. All Rights Reserved.
 *
 * This software, including source code, documentation and related
 * materials ("Software"), is owned by Cypress Semiconductor Corporation
 * or one of its subsidiaries ("Cypress") and is protected by and subject to
 * worldwide patent protection (United States and foreign),
 * United States copyright laws and international treaty provisions.
 * Therefore, you may use this Software only as provided in the license
 * agreement accompanying the software package from which you
 * obtained this Software ("EULA").
 * If no EULA applies, Cypress hereby grants you a personal, non-exclusive,
 * non-transferable license to copy, modify, and compile the Software
 * source code solely for use in connection with Cypress's
 * integrated circuit products. Any reproduction, modification, translation,
 * compilation, or representation of this Software except as specified
 * above is prohibited without the express written permission of Cypress.
 *
 * Disclaimer: THIS SOFTWARE IS PROVIDED AS-IS, WITH NO WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, NONINFRINGEMENT, IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. Cypress
 * reserves the right to make changes to the Software without notice. Cypress
 * does not assume any liability arising out of the application or use of the
 * Software or any product or circuit described in the Software. Cypress does
 * not authorize its products for use in any products where a malfunction or
 * failure of the Cypress product may reasonably be expected to result in
 * significant property damage, injury or death ("High Risk Product"). By
 * including Cypress's product in a High Risk Product, the manufacturer
 * of such system or application assumes all risk of such use and in doing
 * so agrees to indemnify Cypress against all liability.
 */

/*
 * File name: low_power_20819.c
 *
 * Description: Main file which is the entry point to the application.
 *              The application_start function initializes UART for trace
 *              messages, initializes the BT stack and configures sleep
 *              parameters. The low_power_sleep_handler function handles the
 *              sleep related callbacks from the PMU.
 *              Refer to the CE document TODO.pdf
 *              provided along with the code example for more details.
 *
 * Features demonstrated:
 *              SDS and HID-Off functionality with BLE advertisements (ADV) and
 *              connection.
 *
 *
 *  Controls:
 *  - Start ADV with button press when not connected.
 *  - Disconnect and enter HID-Off with button press when connected.
 *  - Wake from HID-Off on button press.
 *
 *  Note: Do not connect HCI UART when running this application as it will
 *  prevent the device from entering sleep as we have configured sleep mode as
 *  WICED_SLEEP_MODE_NO_TRANSPORT (see header file wiced_sleep.h)
 */

#include "app_bt_cfg.h"
#include "sparcommon.h"
#include "wiced_bt_dev.h"
#include "wiced_platform.h"
#include "wiced_bt_trace.h"
#include "wiced_hal_puart.h"
#include "wiced_bt_stack.h"
#include "cycfg_gatt_db.h"
#include "cycfg_pins.h"
#include "wiced_sleep.h"
#include "clock_timer.h"
#include "low_power_20719.h"
#include "spar_utils.h"


/* Notes - To be removed before final submission
 *
 * Remove all Debug related code
 * Change the libraries to latest 3.X (ignore if done)
 * Check host wake and device wake
 * Check TRANSPORT_MODE behaviour
 * Documentation
 * Remove magic numbers
 * Check file headers
 * Check function headers
 * Need instructions in the README to add PLACE_DATA_IN_RETENTION_RAM in front of
 * uint8_t app_bas_battery_level[] and uint8_t app_bas_battery_level_client_char_config[]
 * in the file GeneratedSource/cycfg_gatt_db.c after first compilation and after every
 * clean and build
 * Check if SW3 needs to be replaced by some other macro
 *
 */

/**************************************************************************************
 *                                Variable Definitions
 *************************************************************************************/
/* Maintains the current system state */
PLACE_DATA_IN_RETENTION_RAM uint8_t app_bt_current_state = SLEEP_WITHOUT_BLE;

OSAPI_TIMER app_bt_notification_timer_handle;
PLACE_DATA_IN_RETENTION_RAM uint64_t notification_timer_start_instance;

wiced_sleep_config_t    low_power_sleep_config; /* sleep configuration */

/* Variable used to indicate whether to send notification after fast boot.
 * This is done so that the notification function can be called after
 * stack initialization
 */
uint8_t send_notification_after_sds = 0;

extern uint8_t app_bt_conn_id;

/******************************************************************/

/*******************************************************************
 * Function Definitions
 ******************************************************************/

/*******************************************************************************
* Function Name: void application_start(void)
********************************************************************************
* Summary: Entry point to the application. Initialize transport configuration
*          and register BLE management event callback. The actual application
*          initialization will happen when stack reports that BT device is ready
*
* Parameters:
*   None
*
* Return:
*  None
*
********************************************************************************/
void application_start(void)
{
    osapi_createTimer(&app_bt_notification_timer_handle, &app_bt_timeout, 0);

    /* Select Debug UART setting to see debug traces on the appropriate port */
    wiced_set_debug_uart(WICED_ROUTE_DEBUG_TO_PUART);

    /* Initialize Stack and Register Management Callback */
    wiced_bt_stack_init(app_bt_management_callback, &wiced_bt_cfg_settings,
                                                     wiced_bt_cfg_buf_pools);

    /* configure to sleep if sensor is idle */
    low_power_sleep_config.sleep_mode           = WICED_SLEEP_MODE_NO_TRANSPORT;
    low_power_sleep_config.device_wake_mode     = WICED_SLEEP_WAKE_ACTIVE_LOW;
    low_power_sleep_config.device_wake_source   = WICED_SLEEP_WAKE_SOURCE_GPIO;
    low_power_sleep_config.device_wake_gpio_num = SW3;
    low_power_sleep_config.host_wake_mode       = WICED_SLEEP_WAKE_ACTIVE_HIGH;
    low_power_sleep_config.sleep_permit_handler = app_bt_sleep_handler;

    if(WICED_BT_SUCCESS != wiced_sleep_configure(&low_power_sleep_config))
    {
        WICED_BT_TRACE("Sleep Configure failed\r\n");
    }

    if(wiced_sleep_get_boot_mode())
    {
        WICED_BT_TRACE("Fast boot\n\r");
        uint64_t current_system_time = clock_SystemTimeMicroseconds64();
        uint64_t time_passed_in_ms = (current_system_time - notification_timer_start_instance);

        /* TODO Debug print */
        WICED_BT_TRACE("time_passed_in_ms: 0x%X %X, current_system_time: 0x%X %X,"
        		"notification_timer_start_instance: 0x%X %X\n\r",
				((uint32_t)((time_passed_in_ms & 0xFFFFFFFF00000000) >> 32)), ((uint32_t)(time_passed_in_ms & 0x00000000FFFFFFFF)),
				((uint32_t)((current_system_time & 0xFFFFFFFF00000000) >> 32)), ((uint32_t)(current_system_time & 0x00000000FFFFFFFF)),
				((uint32_t)((notification_timer_start_instance & 0xFFFFFFFF00000000) >> 32)), ((uint32_t)(notification_timer_start_instance & 0x00000000FFFFFFFF)));

        /* TODO Remove magiv number by checking the clock accuracy */
        if((4900000 <= time_passed_in_ms) && (0 < notification_timer_start_instance) && !osapi_is_timer_running(&app_bt_notification_timer_handle))
        {
        	WICED_BT_TRACE("Timer wakeup\n\r");
        	send_notification_after_sds = 1;
        	notification_timer_start_instance = 0;
        }
    }
    else
    {
        WICED_BT_TRACE("Cold boot\n\r");
    }
}

/**************************************************************************************
 * Function Name: uint32_t low_power_sleep_handler(wiced_sleep_poll_type_t type)
 **************************************************************************************
 * Summary: Callback for sleep permissions.
 *
 * Parameters:
 *   wiced_sleep_poll_type_t type          : Poll type (see #wiced_sleep_poll_type_t)
 *
 * Return:
 *   uint32_t: if type == WICED_SLEEP_POLL_SLEEP_PERMISSION, application should return
 *             WICED_SLEEP_ALLOWED_WITHOUT_SHUTDOWN or WICED_SLEEP_NOT_ALLOWED.
 *             if type == WICED_SLEEP_POLL_TIME_TO_SLEEP, application should return
 *             WICED_SLEEP_MAX_TIME_TO_SLEEP
 *
 *************************************************************************************/
uint32_t app_bt_sleep_handler(wiced_sleep_poll_type_t type)
{
    uint32_t ret = WICED_SLEEP_NOT_ALLOWED;

    WICED_BT_TRACE(".");
    switch(type)
    {
        case WICED_SLEEP_POLL_SLEEP_PERMISSION:
            if(SLEEP_WITH_ADV == app_bt_current_state)
            {
                ret = WICED_SLEEP_ALLOWED_WITHOUT_SHUTDOWN;
            }
            else
        	{
        		ret = WICED_SLEEP_ALLOWED_WITH_SHUTDOWN;
        		WICED_BT_TRACE("*");
        	}
            break;

        case WICED_SLEEP_POLL_TIME_TO_SLEEP:
            	ret = WICED_SLEEP_MAX_TIME_TO_SLEEP;
            break;
    }
    return ret;
}

/****************************************************************************
* Function Name: void app_bt_timeout(uint32_t count)
*****************************************************************************
* Summary:
*   Five second timer callback. Send dummy battery value if connected and
*   notification enabled.
*
* Parameters:
*   uint32_t count: Variable registered by wiced_init_timer to pass to
*                    timer callback. Not used here.
*
* Return:
*   None
*
****************************************************************************/
void app_bt_timeout(int32_t count, uint32_t overTimeInUs)
{
	wiced_bt_gatt_status_t notification_status;
	WICED_BT_TRACE("Timer timeout\n\r");
    /* Check if device is connected and notifications are enabled and send
     * notification */
    if((GATT_CLIENT_CONFIG_NOTIFICATION ==
            app_bas_battery_level_client_char_config[0]) &&
            (0 != app_bt_conn_id))
    {
       if(BATTERY_LEVEL_FULL > app_bas_battery_level[0])
       {
           app_bas_battery_level[0]++;
       }
       else
       {
           app_bas_battery_level[0] = BATTERY_LEVEL_EMPTY;
       }

       WICED_BT_TRACE("\r\nSending Notification\r\n");
       WICED_BT_TRACE("Battery level: %d%%\r\n", app_bas_battery_level[0]);
       WICED_BT_TRACE("conn_id: %d, Batt Value HDL: %d, attr_tbl_max_len: %d, "
    		   "attr_tbl_p_data: %d\r\n", app_bt_conn_id, HDLC_BAS_BATTERY_LEVEL_VALUE,
			   app_gatt_db_ext_attr_tbl[2].max_len, *app_gatt_db_ext_attr_tbl[2].p_data);

       notification_status = wiced_bt_gatt_send_notification
    		   	   	   	   	 (app_bt_conn_id, HDLC_BAS_BATTERY_LEVEL_VALUE,
							  app_gatt_db_ext_attr_tbl[2].max_len,
							  app_gatt_db_ext_attr_tbl[2].p_data);

       if(WICED_BT_GATT_SUCCESS != notification_status)
       {
           WICED_BT_TRACE("Sending sensor value notification failed: 0x%X\r\n", notification_status);
       }
       else
       {
    	   notification_timer_start_instance = clock_SystemTimeMicroseconds64();
    	   osapi_activateTimer( &app_bt_notification_timer_handle, NOTFICATION_TIME_MS*1000);

    	   /* TODO Debug print */
    	   WICED_BT_TRACE("notification_timer_start_instance: 0x%X %X, 0x%X\r\n",
    			   ((uint32_t)((notification_timer_start_instance & 0xFFFFFFFF00000000) >> 32)), ((uint32_t)(notification_timer_start_instance & 0x00000000FFFFFFFF)),
				   clock_SystemTimeMicroseconds64());
       }
    }
}
